from collections import defaultdict
from datetime import datetime
from itertools import chain
from math import ceil
from typing import Optional, Iterable, Iterator, List, Dict

from maggma.builders import Builder
from maggma.core import Store
from maggma.utils import grouper

from emmet.core.qchem.task import TaskDocument
from emmet.core.qchem.molecule import MoleculeDoc, evaluate_lot
from emmet.core.molecules.orbitals import OrbitalDoc
from emmet.core.utils import jsanitize
from emmet.builders.settings import EmmetBuildSettings


__author__ = "Evan Spotte-Smith"

SETTINGS = EmmetBuildSettings()


class OrbitalBuilder(Builder):
    """
    The OrbitalBuilder extracts the highest-quality natural bonding orbital data
    from a MoleculeDoc (lowest electronic energy, highest level of theory for
    each solvent available).

    The process is as follows:
        1. Gather MoleculeDocs by formula
        2. For each doc, sort tasks by solvent
        3. For each solvent, grab the best TaskDoc (including NBO data using
            the highest level of theory with lowest electronic energy for the
            molecule)
        4. Convert TaskDoc to OrbitalDoc
    """

    def __init__(
        self,
        tasks: Store,
        molecules: Store,
        orbitals: Store,
        query: Optional[Dict] = None,
        settings: Optional[EmmetBuildSettings] = None,
        **kwargs,
    ):
        self.tasks = tasks
        self.molecules = molecules
        self.orbitals = orbitals
        self.query = query if query else dict()
        self.settings = EmmetBuildSettings.autoload(settings)
        self.kwargs = kwargs

        super().__init__(sources=[tasks, molecules], targets=[orbitals], **kwargs)
        # Uncomment in case of issue with mrun not connecting automatically to collections
        # for i in [self.tasks, self.molecules, self.orbitals]:
        #     try:
        #         i.connect()
        #     except Exception as e:
        #         print("Could not connect,", e)

    def ensure_indexes(self):
        """
        Ensures indices on the collections needed for building
        """

        # Basic search index for tasks
        self.tasks.ensure_index("task_id")
        self.tasks.ensure_index("last_updated")
        self.tasks.ensure_index("state")
        self.tasks.ensure_index("formula_alphabetical")

        # Search index for molecules
        self.molecules.ensure_index("molecule_id")
        self.molecules.ensure_index("last_updated")
        self.molecules.ensure_index("task_ids")
        self.molecules.ensure_index("formula_alphabetical")

        # Search index for orbitals
        self.orbitals.ensure_index("molecule_id")
        self.orbitals.ensure_index("task_id")
        self.orbitals.ensure_index("solvent")
        self.orbitals.ensure_index("lot_solvent")
        self.orbitals.ensure_index("property_id")
        self.orbitals.ensure_index("last_updated")
        self.orbitals.ensure_index("formula_alphabetical")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:  # pragma: no cover
        """Prechunk the builder for distributed computation"""

        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.orbitals.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        N = ceil(len(to_process_forms) / number_splits)

        for formula_chunk in grouper(to_process_forms, N):
            yield {"query": {"formula_alphabetical": {"$in": list(formula_chunk)}}}

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets all items to process into orbital documents.
        This does no datetime checking; relying on on whether
        task_ids are included in the orbitals Store

        Returns:
            generator or list relevant tasks and molecules to process into documents
        """

        self.logger.info("Orbital builder started")
        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Save timestamp to mark buildtime
        self.timestamp = datetime.utcnow()

        # Get all processed molecules
        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.orbitals.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        self.logger.info(f"Found {len(to_process_docs)} unprocessed documents")
        self.logger.info(f"Found {len(to_process_forms)} unprocessed formulas")

        # Set total for builder bars to have a total
        self.total = len(to_process_forms)

        for formula in to_process_forms:
            mol_query = dict(temp_query)
            mol_query["formula_alphabetical"] = formula
            molecules = list(self.molecules.query(criteria=mol_query))

            yield molecules

    def process_item(self, items: List[Dict]) -> List[Dict]:
        """
        Process the tasks into a OrbitalDocs

        Args:
            tasks List[Dict] : a list of MoleculeDocs in dict form

        Returns:
            [dict] : a list of new orbital docs
        """

        mols = [MoleculeDoc(**item) for item in items]
        formula = mols[0].formula_alphabetical
        mol_ids = [m.molecule_id for m in mols]
        self.logger.info(f"Processing {formula} : {mol_ids}")

        orbital_docs = list()

        for mol in mols:
            correct_charge_spin = [
                e
                for e in mol.entries
                if e["charge"] == mol.charge
                and e["spin_multiplicity"] == mol.spin_multiplicity
            ]

            # Must have NBO, and must specifically use NBO7
            orbital_entries = [
                e
                for e in correct_charge_spin
                if e["output"]["nbo"] is not None
                and (
                    e["orig"]["rem"].get("run_nbo6", False)
                    or e["orig"]["rem"].get("nbo_external", False)
                )
            ]

            # Organize by solvent environment
            by_solvent = defaultdict(list)
            for entry in orbital_entries:
                by_solvent[entry["solvent"]].append(entry)

            for solvent, entries in by_solvent.items():
                # No documents with NBO data; no documents to be made
                if len(entries) == 0:
                    continue
                else:
                    sorted_entries = sorted(
                        entries,
                        key=lambda x: (
                            sum(evaluate_lot(x["level_of_theory"])),
                            x["energy"],
                        ),
                    )

                    for best in sorted_entries:
                        task = best["task_id"]

                        tdoc = self.tasks.query_one(
                            {
                                "task_id": task,
                                "formula_alphabetical": formula,
                                "orig": {"$exists": True},
                            }
                        )

                        if tdoc is None:
                            try:
                                tdoc = self.tasks.query_one(
                                    {
                                        "task_id": int(task),
                                        "formula_alphabetical": formula,
                                        "orig": {"$exists": True},
                                    }
                                )
                            except ValueError:
                                tdoc = None

                        if tdoc is None:
                            continue

                        task_doc = TaskDocument(**tdoc)

                        if task_doc is None:
                            continue

                        orbital_doc = OrbitalDoc.from_task(
                            task_doc, molecule_id=mol.molecule_id, deprecated=False
                        )

                        if orbital_doc is not None:
                            orbital_docs.append(orbital_doc)

        self.logger.debug(f"Produced {len(orbital_docs)} orbital docs for {formula}")

        return jsanitize([doc.model_dump() for doc in orbital_docs], allow_bson=True)

    def update_targets(self, items: List[List[Dict]]):
        """
        Inserts the new documents into the orbitals collection

        Args:
            items [[dict]]: A list of documents to update
        """

        docs = list(chain.from_iterable(items))  # type: ignore

        # Add timestamp
        for item in docs:
            item.update(
                {
                    "_bt": self.timestamp,
                }
            )

        molecule_ids = list({item["molecule_id"] for item in docs})

        if len(items) > 0:
            self.logger.info(f"Updating {len(docs)} orbital documents")
            self.orbitals.remove_docs({self.orbitals.key: {"$in": molecule_ids}})
            self.orbitals.update(
                docs=docs,
                key=["molecule_id", "solvent"],
            )
        else:
            self.logger.info("No items to update")



from itertools import combinations
import itertools
from math import ceil
from typing import Dict, List, Iterator, Optional

from maggma.core import Builder
from maggma.stores import S3Store, MongoURIStore, MongoStore
from maggma.utils import grouper
from pymatgen.core import Composition, Element


class MissingCompositionsBuilder(Builder):
    """
    Builder that finds compositions not found in the
    Materials Project for each chemical system.
    Based on the Text Mining project in MPContribs.
    """

    def __init__(
        self,
        phase_diagram: S3Store,
        mpcontribs: MongoURIStore,
        missing_compositions: MongoStore,
        query: Optional[Dict] = None,
        **kwargs,
    ):
        """
        Arguments:
            phase_diagram: source store for chemsys data
            matsholar_store: source store for matscholar data
            missing_compositions: Target store to save the missing compositions
            query: dictionary to query the phase diagram store
            **kwargs: Additional keyword arguments
        """
        self.phase_diagram = phase_diagram
        self.mpcontribs = mpcontribs
        self.missing_compositions = missing_compositions
        self.query = query
        self.kwargs = kwargs
        # TODO: make sure the two lines below are needed?
        self.phase_diagram.key = "phase_diagram_id"
        self.missing_compositions.key = "chemical_system"

        super().__init__(
            sources=[phase_diagram, mpcontribs],
            targets=[missing_compositions],
            **kwargs,
        )

    def prechunk(self, number_splits: int) -> Iterator[Dict]:  # pragma: no cover
        """
        Prechunk method to perform chunking by the key field
        """
        q = self.query or {}  # type: ignore

        keys = self.missing_compositions.newer_in(
            self.phase_diagram, criteria=q, exhaustive=True
        )

        N = ceil(len(keys) / number_splits)
        for split in grouper(keys, N):
            yield {"query": {self.phase_diagram.key: {"$in": list(split)}}}

    def get_items(self) -> Iterator[Dict]:
        """
        Returns all chemical systems (combinations of elements)
        to process.
        Enumarates all chemical systems and queries the
        phase diagram for each system, in the case where
        the chemical system is not found in the phase diagram,
        it returns a dictionary with the chemical system
        and an empty list for the missing compositions
        """
        self.logger.info("Missing Composition Builder Started")
        self.logger.info("Setting up chemical systems to process")
        elements = set()
        # get all elements
        elements = set([e.symbol for e in Element])

        # Generate all unique combinations of elements to form chemical systems
        chemical_systems = []
        for r in range(2, 5):
            for combination in combinations(elements, r):
                system = "-".join(sorted([str(element) for element in combination]))
                chemical_systems.append(system)
        q = self.query or {}
        projection = {
            "chemsys": 1,
            "phase_diagram.all_entries.composition": 1,
        }
        for sys in chemical_systems:
            q.update({"chemsys": sys})
            self.logger.info(f"Querying phase diagram for {sys}")
            try:
                items = self.phase_diagram.query(criteria=q, properties=projection)
                # combine all entries from all phase diagrams
                all_entries = []
                for item in items:
                    all_entries = [
                        i["composition"] for i in item["phase_diagram"]["all_entries"]
                    ]

                # Find missing compositions
                matscholar_entries = self._get_entries_in_chemsys(sys)
                doc = {
                    "chemsys": sys,
                    "all_compositions": all_entries,
                    "matscholar_entries": matscholar_entries,
                }
                yield doc
            except Exception as ex:
                self.logger.error(f"Erro looking for phase diagram for {sys}: {ex}")
                continue

    def process_item(self, item: Dict) -> Dict:
        """
        Processes a chemical system and finds missing c
        ompositions for that system.
        Note that it returns a missing_compositions dict
        regardless of whether there is a missing composition,
        in which case, it contains an empty dictionary for
        the missing_composition_entries field.
        """
        compositions = set()
        chemsys = item["chemsys"]
        matscholar_entries = item["matscholar_entries"]
        self.logger.info(
            "Querying entries in MPContribs matscholar"
            f"project for the chemical system {chemsys}"
        )
        missing_compositions = {
            "chemical_system": chemsys,
            "missing_composition_entries": {},
        }

        if len(item["all_compositions"]) > 0:
            # Get the compositions from retrieved entries,
            # and use its reduced_formula
            for entry in item["all_compositions"]:
                composition = Composition.from_dict(entry)
                # Note the reduced formula is a string
                # instead of a Composition object
                compositions.add(composition.reduced_formula)

            if len(matscholar_entries) == 0:
                self.logger.info(
                    "No entries found in MPContribs" "for the chemical system"
                )

            else:
                self.logger.info(
                    f"Found {len(matscholar_entries)}"
                    "entries in MPContribs for the chemical system"
                )

                for entry in matscholar_entries:
                    # Comparing reduced formulae from MPContribs
                    # and Phase Diagram
                    if (
                        Composition(entry["formula"]).reduced_formula
                        not in compositions
                    ):
                        # this formula doesn't exist in the dictionary,
                        # make an entry in the missing_compositions dict
                        if (
                            entry["formula"]
                            not in missing_compositions[
                                "missing_composition_entries"
                            ].keys()
                        ):
                            missing_compositions["missing_composition_entries"].update(
                                {
                                    entry["formula"]: [
                                        {"link": entry["link"], "doi": entry["doi"]}
                                    ]
                                }
                            )
                        # formula already exists in the dictionary, append the new entry
                        else:
                            missing_compositions["missing_composition_entries"][
                                entry["formula"]
                            ].append({"link": entry["link"], "doi": entry["doi"]})

        return missing_compositions

    def update_targets(self, items):
        """
        Updates the target store with the missing compositions
        """
        docs = list(filter(None, items))

        if len(docs) > 0:
            self.logger.info(f"Found {len(docs)} chemical-system docs to update")
            self.missing_compositions.update(items)
        else:
            self.logger.info("No items to update")

    def _get_entries_in_chemsys(self, chemsys) -> List:
        """Queries the MPContribs Store for entries in a chemical system."""
        # get sub-systems
        chemsys_subsystems = []
        elements = chemsys.split("-")
        # get all possible combinations
        for i in range(2, len(elements) + 1):
            chemsys_subsystems += [
                "-".join(sorted(c)) for c in itertools.combinations(elements, i)
            ]

        results = []
        for subsystem in chemsys_subsystems:
            try:
                query = {"project": "matscholar", "data.chemsys": subsystem}
                fields = ["formula", "data"]
                entries = self.mpcontribs.query(criteria=query, properties=fields)
                for entry in entries:
                    results.append({"formula": entry["formula"], **entry["data"]})
            except Exception as ex:
                self.logger.error(f"Error querying MPContribs for {subsystem}: {ex}")
        return results



from datetime import datetime
from itertools import chain
from math import ceil
from typing import Optional, Iterable, Iterator, List, Dict
import copy

from pymatgen.core.structure import Molecule
from pymatgen.util.graph_hashing import weisfeiler_lehman_graph_hash

from maggma.builders import Builder
from maggma.core import Store
from maggma.utils import grouper

from emmet.core.qchem.molecule import MoleculeDoc
from emmet.core.molecules.atomic import PartialChargesDoc, PartialSpinsDoc
from emmet.core.molecules.bonds import MoleculeBondingDoc, metals
from emmet.core.molecules.thermo import MoleculeThermoDoc
from emmet.core.molecules.metal_binding import MetalBindingDoc, METAL_BINDING_METHODS
from emmet.core.utils import jsanitize
from emmet.builders.settings import EmmetBuildSettings


__author__ = "Evan Spotte-Smith"

SETTINGS = EmmetBuildSettings()


class MetalBindingBuilder(Builder):
    """
    The MetalBindingBuilder extracts information about metal binding in molecules
    from MoleculeDocs, MoleculeThermoDocs, MoleculeBondingDocs, PartialChargesDocs,
    and PartialSpinsDocs.

    NBO is the strongly preferred method to approximate partial charges and bonding,
    and so if NBO bonding and partial charges/spins documents are available, they will
    be used.

    If NBO docs are not available, then bonding can be taken from the "OpenBabelNN + metal_edge_extender"
    method, and partial charges and spins can be taken from "mulliken".

    This builder will attempt to build documents for each molecule in each solvent.
    For each molecule-solvent combination, the highest-quality data available
    (based on level of theory, electronic energy, and the method used to generate bonding/charges/spins)
    will be used.

    The process is as follows:
        1. Gather MoleculeDocs by formula
        2. For each molecule, first identify if there are any metals. If not, then no MetalBindingDoc can be made.
            If so, then identify the possible solvents that can be used to generate MetalBindingDocs
        3. For each combination of Molecule ID and solvent, search for additional documents:
            - MoleculeBondingDocs
            - PartialChargesDocs
            - PartialSpinsDocs (for open-shell molecules)
            - MoleculeThermoDocs
        4. Group these additional documents by level of theory and (where applicable) method, and choose the best
            possible methods for which all required data is available
        5. For each metal in the molecule:
            5.1 Use partial charge and spin information to determine the oxidation and spin state of the metal
            5.2 Search for MoleculeThermoDocs for the metal atom/ion with appropriate charge and spin with the
                chosen level of theory
            5.3 Use graph comparisons (hashing or isomorphism) to identify a molecule with the same structure as
                the molecule of interest WITHOUT the metal of interest, as well as the appropriate charge and spin
            5.4 If an appropriate metal-less molecule can be found, search for a MoleculeThermoDoc for that molecule'
                with the chosen level of theory
        6. Use the obtained bonding, charges, spins, and thermo docs to construct a MetalBindingDoc
    """

    def __init__(
        self,
        molecules: Store,
        charges: Store,
        spins: Store,
        bonds: Store,
        thermo: Store,
        metal_binding: Store,
        query: Optional[Dict] = None,
        methods: Optional[List] = None,
        settings: Optional[EmmetBuildSettings] = None,
        **kwargs,
    ):
        self.molecules = molecules
        self.charges = charges
        self.spins = spins
        self.bonds = bonds
        self.thermo = thermo
        self.metal_binding = metal_binding
        self.query = query if query else dict()
        self.methods = methods if methods else METAL_BINDING_METHODS
        self.settings = EmmetBuildSettings.autoload(settings)
        self.kwargs = kwargs

        super().__init__(
            sources=[molecules, charges, spins, bonds, thermo],
            targets=[metal_binding],
            **kwargs,
        )
        # Uncomment in case of issue with mrun not connecting automatically to collections
        # for i in [self.molecules, self.charges, self.spins, self.bonds, self.thermo, self.metal_binding]:
        #     try:
        #         i.connect()
        #     except Exception as e:
        #         print("Could not connect,", e)

    def ensure_indexes(self):
        """
        Ensures indices on the collections needed for building
        """

        # Search index for molecules
        self.molecules.ensure_index("molecule_id")
        self.molecules.ensure_index("last_updated")
        self.molecules.ensure_index("task_ids")
        self.molecules.ensure_index("formula_alphabetical")

        # Search index for charges
        self.charges.ensure_index("molecule_id")
        self.charges.ensure_index("task_id")
        self.charges.ensure_index("method")
        self.charges.ensure_index("solvent")
        self.charges.ensure_index("lot_solvent")
        self.charges.ensure_index("property_id")
        self.charges.ensure_index("last_updated")
        self.charges.ensure_index("formula_alphabetical")

        # Search index for spins
        self.spins.ensure_index("molecule_id")
        self.spins.ensure_index("task_id")
        self.spins.ensure_index("method")
        self.spins.ensure_index("solvent")
        self.spins.ensure_index("lot_solvent")
        self.spins.ensure_index("property_id")
        self.spins.ensure_index("last_updated")
        self.spins.ensure_index("formula_alphabetical")

        # Search index for bonds
        self.bonds.ensure_index("molecule_id")
        self.bonds.ensure_index("method")
        self.bonds.ensure_index("task_id")
        self.bonds.ensure_index("solvent")
        self.bonds.ensure_index("lot_solvent")
        self.bonds.ensure_index("property_id")
        self.bonds.ensure_index("last_updated")
        self.bonds.ensure_index("formula_alphabetical")

        # Search index for thermo
        self.thermo.ensure_index("molecule_id")
        self.thermo.ensure_index("task_id")
        self.thermo.ensure_index("solvent")
        self.thermo.ensure_index("lot_solvent")
        self.thermo.ensure_index("property_id")
        self.thermo.ensure_index("last_updated")
        self.thermo.ensure_index("formula_alphabetical")

        # Search index for metal_binding
        self.metal_binding.ensure_index("molecule_id")
        self.metal_binding.ensure_index("solvent")
        self.metal_binding.ensure_index("lot_solvent")
        self.metal_binding.ensure_index("property_id")
        self.metal_binding.ensure_index("last_updated")
        self.metal_binding.ensure_index("formula_alphabetical")
        self.metal_binding.ensure_index("method")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:  # pragma: no cover
        """Prechunk the builder for distributed computation"""

        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.metal_binding.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        N = ceil(len(to_process_forms) / number_splits)

        for formula_chunk in grouper(to_process_forms, N):
            yield {"query": {"formula_alphabetical": {"$in": list(formula_chunk)}}}

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets all items to process into metal_binding documents.

        Returns:
            generator or list relevant molecules to process into documents
        """

        self.logger.info("Metal binding builder started")
        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Save timestamp to mark buildtime
        self.timestamp = datetime.utcnow()

        # Get all processed molecules
        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.metal_binding.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        self.logger.info(f"Found {len(to_process_docs)} unprocessed documents")
        self.logger.info(f"Found {len(to_process_forms)} unprocessed formulas")

        # Set total for builder bars to have a total
        self.total = len(to_process_forms)

        for formula in to_process_forms:
            mol_query = dict(temp_query)
            mol_query["formula_alphabetical"] = formula
            molecules = list(self.molecules.query(criteria=mol_query))

            yield molecules

    def process_item(self, items: List[Dict]) -> List[Dict]:
        """
        Process molecule, bonding, partial charges, partial spins, and thermo documents into MetalBindingDocs

        Args:
            tasks List[Dict] : a list of MoleculeDocs in dict form

        Returns:
            [dict] : a list of new metal binding docs
        """

        mols = [MoleculeDoc(**item) for item in items]
        formula = mols[0].formula_alphabetical
        mol_ids = [m.molecule_id for m in mols]
        self.logger.debug(f"Processing {formula} : {mol_ids}")

        binding_docs = list()

        for mol in mols:
            # First: do we need to do this? Are there actually metals in this molecule? And species other than metals?
            species = mol.species
            metal_indices = [i for i, e in enumerate(species) if e in metals]
            if len(metal_indices) == 0 or len(species) == 1:
                # print(mol.molecule_id, mol.formula_alphabetical)
                continue

            # Grab the basic documents needed to create a metal binding document
            molecule_id = mol.molecule_id
            solvents = mol.unique_solvents
            charges = [
                PartialChargesDoc(**e)
                for e in self.charges.query({"molecule_id": molecule_id})
            ]
            if mol.spin_multiplicity != 1:
                spins = [
                    PartialSpinsDoc(**e)
                    for e in self.spins.query({"molecule_id": molecule_id})
                ]
            else:
                spins = list()
            bonds = [
                MoleculeBondingDoc(**e)
                for e in self.bonds.query({"molecule_id": molecule_id})
            ]
            thermo = [
                MoleculeThermoDoc(**e)
                for e in self.thermo.query({"molecule_id": molecule_id})
            ]

            if any([len(x) == 0 for x in [charges, bonds, thermo]]):
                # Not enough information to construct MetalBindingDoc
                continue
            elif mol.spin_multiplicity != 1 and len(spins) == 0:
                # For open-shell molecule, partial spins information needed
                continue

            # Group by solvent and (where appropriate) method
            charge_bysolv_meth = dict()
            for c in charges:
                if c.solvent not in charge_bysolv_meth:
                    charge_bysolv_meth[c.solvent] = {c.method: c}
                else:
                    charge_bysolv_meth[c.solvent][c.method] = c

            spins_bysolv_meth = dict()
            for s in spins:
                if s.solvent not in spins_bysolv_meth:
                    spins_bysolv_meth[s.solvent] = {s.method: s}
                else:
                    spins_bysolv_meth[s.solvent][s.method] = s

            bonds_bysolv_meth = dict()
            for b in bonds:
                if b.solvent not in bonds_bysolv_meth:
                    bonds_bysolv_meth[b.solvent] = {b.method: b}
                else:
                    bonds_bysolv_meth[b.solvent][b.method] = b

            thermo_bysolv = {t.solvent: t for t in thermo}

            for solvent in solvents:
                this_charge = charge_bysolv_meth.get(solvent)  # type: ignore
                this_spin = spins_bysolv_meth.get(solvent)  # type: ignore
                this_bond = bonds_bysolv_meth.get(solvent)  # type: ignore
                base_thermo_doc = thermo_bysolv.get(solvent)  # type: ignore

                # Do we have the requisite docs for this solvent?
                if mol.spin_multiplicity == 1:
                    needed = [this_charge, this_bond, base_thermo_doc]
                else:
                    needed = [this_charge, this_spin, this_bond, base_thermo_doc]

                if any([x is None for x in needed]):
                    continue

                # What method will we use?
                # Currently allows two options:
                # 1. Using NBO for everything ("nbo")
                # 2. Using Mulliken for charges/spins and OpenBabel + metal_edge_extender for bonding
                #   ("mulliken-OB-mee")
                for method in self.methods:
                    plan = False
                    if mol.spin_multiplicity == 1:
                        if method == "nbo" and all(
                            [x.get("nbo") is not None for x in [this_charge, this_bond]]  # type: ignore
                        ):  # type: ignore
                            plan = True
                            charge_doc = this_charge.get("nbo")  # type: ignore
                            spin_doc = None
                            bond_doc = this_bond.get("nbo")  # type: ignore
                        elif method == "mulliken-OB-mee" and (
                            this_charge.get("mulliken") is not None  # type: ignore
                            and this_bond.get("OpenBabelNN + metal_edge_extender") is not None  # type: ignore
                        ):
                            plan = True
                            charge_doc = this_charge.get("mulliken")  # type: ignore
                            spin_doc = None
                            bond_doc = this_bond.get("OpenBabelNN + metal_edge_extender")  # type: ignore
                    else:
                        if method == "nbo" and all(
                            [x.get("nbo") is not None for x in [this_charge, this_spin, this_bond]]  # type: ignore
                        ):  # type: ignore
                            charge_lot = this_charge.get("nbo").level_of_theory  # type: ignore
                            spin_lot = this_spin.get("nbo").level_of_theory  # type: ignore
                            if charge_lot == spin_lot:  # type: ignore
                                plan = True
                                charge_doc = this_charge.get("nbo")  # type: ignore
                                spin_doc = this_spin.get("nbo")  # type: ignore
                                bond_doc = this_bond.get("nbo")  # type: ignore
                        elif (
                            method == "mulliken-OB-mee"
                            and this_charge.get("mulliken") is not None  # type: ignore
                            and this_spin.get("mulliken") is not None  # type: ignore
                            and this_bond.get("OpenBabelNN + metal_edge_extender") is not None  # type: ignore
                        ):
                            charge_lot = this_charge.get("mulliken").level_of_theory  # type: ignore
                            spin_lot = this_spin.get("mulliken").level_of_theory  # type: ignore
                            if charge_lot == spin_lot:  # type: ignore
                                plan = True
                                charge_doc = this_charge.get("mulliken")  # type: ignore
                                spin_doc = this_spin.get("mulliken")  # type: ignore
                                bond_doc = this_bond.get("OpenBabelNN + metal_edge_extender")  # type: ignore

                    # Don't have the right combinations of level of theory and method
                    if plan is False:
                        continue

                    # Obtain relevant thermo documents for each metal atom/ion in the molecule
                    metal_thermo = dict()
                    nometal_thermo = dict()
                    for metal_index in metal_indices:
                        # First, determine the appropriate charge and spin of the metal
                        # TODO: figure out better charge assignment
                        partial_charge = charge_doc.partial_charges[metal_index]  # type: ignore

                        if mol.spin_multiplicity == 1:
                            # For now, just round to nearest whole number
                            charge = round(partial_charge)
                            spin = 1
                        else:
                            partial_spin = spin_doc.partial_spins[metal_index]  # type: ignore
                            charge = round(partial_charge)
                            spin = round(abs(partial_spin)) + 1

                        # Sanity check that charge and spin are compatible
                        metal_species = species[metal_index]
                        try:
                            _ = Molecule(
                                [metal_species],
                                [[0.0, 0.0, 0.0]],
                                charge=charge,
                                spin_multiplicity=spin,
                            )
                        except ValueError:
                            # Assume spin assignment is correct, and change charge accordingly
                            diff_up = abs(partial_charge - (charge + 1))
                            diff_down = abs(partial_charge - (charge - 1))
                            if diff_up < diff_down:
                                charge += 1
                            else:
                                charge -= 1

                        # Grab thermo doc for the relevant metal ion/atom (if available)
                        this_metal_thermo = [
                            MoleculeThermoDoc(**e)
                            for e in self.thermo.query(
                                {
                                    "formula_alphabetical": f"{metal_species}1",
                                    "charge": charge,
                                    "spin_multiplicity": spin,
                                    "lot_solvent": base_thermo_doc.lot_solvent,  # type: ignore
                                }
                            )
                        ]
                        if len(this_metal_thermo) == 0:
                            continue

                        this_metal_thermo = this_metal_thermo[0]
                        metal_thermo[metal_index] = this_metal_thermo

                        # Now the (somewhat) harder part - finding the document for this molecule without the metal
                        # Make sure charges and spins add up
                        nometal_charge = mol.charge - charge
                        nometal_spin = mol.spin_multiplicity - spin + 1
                        mg_copy = copy.deepcopy(bond_doc.molecule_graph)  # type: ignore
                        mg_copy.remove_nodes([metal_index])
                        new_hash = weisfeiler_lehman_graph_hash(
                            mg_copy.graph.to_undirected(), node_attr="specie"
                        )
                        nometal_mol_doc = [
                            MoleculeDoc(**e)
                            for e in self.molecules.query(
                                {
                                    "species_hash": new_hash,
                                    "charge": nometal_charge,
                                    "spin_multiplicity": nometal_spin,
                                }
                            )
                        ]
                        if len(nometal_mol_doc) == 0:
                            continue

                        nometal_mol_id = nometal_mol_doc[0].molecule_id
                        this_nometal_thermo = [
                            MoleculeThermoDoc(**e)
                            for e in self.thermo.query(
                                {
                                    "molecule_id": nometal_mol_id,
                                    "lot_solvent": base_thermo_doc.lot_solvent,  # type: ignore
                                }
                            )
                        ]
                        if len(this_nometal_thermo) == 0:
                            continue

                        this_nometal_thermo = this_nometal_thermo[0]
                        nometal_thermo[metal_index] = this_nometal_thermo

                    doc = MetalBindingDoc.from_docs(
                        method=method,
                        metal_indices=metal_indices,
                        base_molecule_doc=mol,
                        partial_charges=charge_doc,
                        partial_spins=spin_doc,
                        bonding=bond_doc,
                        base_thermo=base_thermo_doc,
                        metal_thermo=metal_thermo,
                        nometal_thermo=nometal_thermo,
                    )

                    if doc is not None and len(doc.binding_data) != 0:
                        binding_docs.append(doc)

        self.logger.debug(
            f"Produced {len(binding_docs)} metal binding docs for {formula}"
        )

        return jsanitize([doc.model_dump() for doc in binding_docs], allow_bson=True)

    def update_targets(self, items: List[List[Dict]]):
        """
        Inserts the new documents into the metal_binding collection

        Args:
            items [[dict]]: A list of documents to update
        """

        docs = list(chain.from_iterable(items))  # type: ignore

        # Add timestamp
        for item in docs:
            item.update(
                {
                    "_bt": self.timestamp,
                }
            )

        molecule_ids = list({item["molecule_id"] for item in docs})

        if len(items) > 0:
            self.logger.info(f"Updating {len(docs)} metal binding documents")
            self.metal_binding.remove_docs(
                {self.metal_binding.key: {"$in": molecule_ids}}
            )
            # Neither molecule_id nor solvent need to be unique, but the combination must be
            self.metal_binding.update(
                docs=docs,
                key=["molecule_id", "solvent", "method"],
            )
        else:
            self.logger.info("No items to update")



from typing import Set, Union, Any
import sys
import os
from gzip import GzipFile
import orjson
import json
from io import BytesIO
from monty.serialization import MontyDecoder
from botocore.exceptions import ClientError
from itertools import chain, combinations
from pymatgen.core import Structure
from pymatgen.analysis.diffusion.neb.full_path_mapper import MigrationGraph


def maximal_spanning_non_intersecting_subsets(sets) -> Set[Set]:
    """
    Finds the maximal spanning non intersecting subsets of a group of sets
    This is usefull for parsing out the sandboxes and figuring out how to group
    and calculate these for thermo documents

    sets (set(frozenset)): sets of keys to subsect, expected as a set of frozensets
    """
    to_return_subsets = []

    # Find the overlapping portions and independent portions
    for subset in sets:
        for other_set in sets:
            subset = frozenset(subset.intersection(other_set)) or subset
        if subset:
            to_return_subsets.append(subset)

    # Remove accounted for elements and recurse on remaining sets
    accounted_elements = set(chain.from_iterable(to_return_subsets))
    sets = {frozenset(subset - accounted_elements) for subset in sets}
    sets = {subset for subset in sets if subset}

    if sets:
        to_return_subsets.extend(maximal_spanning_non_intersecting_subsets(sets))

    return set(to_return_subsets)


def chemsys_permutations(chemsys) -> Set:
    # Function to get all relevant chemical subsystems
    # e.g. for Li-Mn-O returns Li, Li-Mn, Li-Mn-O, Li-O, Mn, Mn-O, O
    elements = chemsys.split("-")
    return {
        "-".join(sorted(c))
        for c in chain(
            *[combinations(elements, i) for i in range(1, len(elements) + 1)]
        )
    }


def get_hop_cutoff(
    migration_graph_struct: Structure,
    mobile_specie: str,
    algorithm: str = "min_distance",
    min_hop_distance: float = 1,
    max_hop_distance: float = 7,
) -> Union[float, None]:
    """
    A function to get an appropriate hop distance cutoff for a given migration
    graph structure which can be used for MigrationGraph.with_distance()

    migration_graph_struct (Structure): the host structure with all working ion
        sites for consideration filled.
        Can get via MigrationGraph.get_structure_from_entries()
    mobile_specie (str): specifies the mobile specie in migration_graph_struct
        (e.g. "Li", "Mg").
    algorithm (str): specify the algorithm to use for getting the hop cutoff.
        "min_distance" =  incrementally increases the hop cutoff to get the
            minimum cutoff value that results in finding a path
        "hops_based" = incrementally increases the hop cutoff until either
            1) the largest hop length exceeds 2x the smallest hop length or
            2) the next incremental increase of the cutoff results in the total
            number of unique hops to exceeding more than 1.5x the previous
            number of unique hops
    """
    d = min_hop_distance
    mg = MigrationGraph.with_distance(
        structure=migration_graph_struct,
        migrating_specie=mobile_specie,
        max_distance=d,
    )
    mg.assign_cost_to_graph(["hop_distance"])
    paths = list(mg.get_path())

    if algorithm == "min_distance":
        while len(paths) == 0 and d < max_hop_distance:
            d = d * 1.1
            mg = MigrationGraph.with_distance(
                structure=migration_graph_struct,
                migrating_specie=mobile_specie,
                max_distance=d,
            )
            mg.assign_cost_to_graph(["hop_distance"])
            paths = list(mg.get_path())
        if d > max_hop_distance:
            return 0
        else:
            return d
    elif algorithm == "hops_based":
        # Thank you to Alex Nguyen Ngoc (alexnguyen0512) for their initial
        # work in developing this algorithm

        # incrementally increase d until at least one path is found
        # this sets d at the minimum distance that results in a path
        while len(paths) == 0 and d < max_hop_distance:
            d = d * 1.1
            mg = MigrationGraph.with_distance(
                structure=migration_graph_struct,
                migrating_specie=mobile_specie,
                max_distance=d,
            )
            mg.assign_cost_to_graph(["hop_distance"])
            paths = list(mg.get_path())
        if d > max_hop_distance:
            return 0

        # once a path is found, get the smallest hop distance and
        # the number of unique hops
        smallest_hop_distance = min(
            [v["hop_distance"] for v in mg.unique_hops.values()]
        )
        num_unique_hops = len(mg.unique_hops)

        # continually incrementally increase d until either...
        # 1) the largest hop distance exceeds 2x the smallest hop distance
        # 2) the number of unique hops spikes (increases by more than 1.5x)
        while (
            max([v["hop_distance"] for v in mg.unique_hops.values()])
            <= smallest_hop_distance * 2
            and d < max_hop_distance
        ):
            d = d * 1.1
            mg = MigrationGraph.with_distance(
                structure=migration_graph_struct,
                migrating_specie=mobile_specie,
                max_distance=d,
            )

            # check for spike in number of hops
            if len(mg.unique_hops) >= 1.5 * num_unique_hops and num_unique_hops >= 4:
                return d / 1.1  # return value of d before the number of hops spiked

            num_unique_hops = len(mg.unique_hops)

        return d

    else:
        return None


def query_open_data(
    bucket: str,
    prefix: str,
    key: str,
    monty_decode: bool = True,
    s3_resource: Any = None,
) -> Union[dict, None]:
    """Query a Materials Project AWS S3 Open Data bucket directly with boto3

    Args:
        bucket (str): Materials project bucket name
        prefix (str): Full set of file prefixes
        key (str): Key for file
        monty_decode (bool): Whether to monty decode or keep as dictionary. Defaults to True.
        s3_resource (Optional[Any]): S3 resource. One will be instantiated if none are provided

    Returns:
        dict: MontyDecoded data or None
    """

    def decode(content, monty_decode):
        if monty_decode:
            result = MontyDecoder().decode(content)
        else:
            result = orjson.loads(content)
        return result

    try:
        file_key = f"{prefix}/{key}.json.gz"
        ref = s3_resource.Object(bucket, file_key)  # type: ignore
        bytes = ref.get()["Body"]  # type: ignore

        with GzipFile(fileobj=bytes, mode="r") as gzipfile:
            content = gzipfile.read()

        try:
            result = decode(content, monty_decode)
        except (orjson.JSONDecodeError, json.JSONDecodeError):
            try:
                with GzipFile(fileobj=BytesIO(content), mode="r") as gzipfile_nested:
                    result = decode(gzipfile_nested.read(), monty_decode)
            except Exception:
                print(f"Issue decoding {file_key} from bucket {bucket}")
                return None
    except ClientError:
        return None

    return result


# From: https://stackoverflow.com/a/45669280
class HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout



from pathlib import Path

import pytest
from maggma.stores import JSONStore, MemoryStore
from monty.serialization import dumpfn, loadfn

from emmet.builders.materials.electronic_structure import ElectronicStructureBuilder
from emmet.builders.vasp.materials import MaterialsBuilder


@pytest.fixture(scope="session")
def tasks_store(test_dir):
    return JSONStore(
        test_dir / "electronic_structure/es_task_docs.json.gz", key="task_id"
    )


@pytest.fixture(scope="session")
def materials_store(tasks_store):
    materials_store = MemoryStore(key="material_id")
    builder = MaterialsBuilder(tasks=tasks_store, materials=materials_store)
    builder.run()
    return materials_store


@pytest.fixture
def electronic_structure_store():
    return MemoryStore(key="material_id")


@pytest.fixture
def bandstructure_fs(test_dir):
    return JSONStore(
        test_dir / "electronic_structure/es_bs_objs.json.gz", key="task_id"
    )


@pytest.fixture
def dos_fs(test_dir):
    return JSONStore(
        test_dir / "electronic_structure/es_dos_objs.json.gz", key="task_id"
    )


def test_electronic_structure_builder(
    tasks_store, materials_store, electronic_structure_store, bandstructure_fs, dos_fs
):
    builder = ElectronicStructureBuilder(
        tasks=tasks_store,
        materials=materials_store,
        electronic_structure=electronic_structure_store,
        bandstructure_fs=bandstructure_fs,
        dos_fs=dos_fs,
    )

    builder.run()
    assert electronic_structure_store.count() == 3


def test_serialization(tmpdir):
    builder = ElectronicStructureBuilder(
        MemoryStore(), MemoryStore(), MemoryStore(), MemoryStore(), MemoryStore()
    )

    dumpfn(builder.as_dict(), Path(tmpdir) / "test.json")
    loadfn(Path(tmpdir) / "test.json")



from math import ceil
from typing import Dict, Iterator, Optional

from maggma.builders import Builder
from maggma.core import Store
from maggma.utils import grouper
from pymatgen.core.structure import Structure

from emmet.core.optimade import OptimadeMaterialsDoc
from emmet.core.utils import jsanitize


class OptimadeMaterialsBuilder(Builder):
    def __init__(
        self,
        materials: Store,
        thermo: Store,
        optimade: Store,
        query: Optional[Dict] = None,
        **kwargs,
    ):
        """
        Creates Optimade compatible docs containing structure and thermo data for materials

        Args:
            materials: Store of materials docs
            thermo: Store of thermo docs
            optimade: Store to update with optimade document
            query : query on materials to limit search
        """

        self.materials = materials
        self.thermo = thermo
        self.optimade = optimade
        self.query = query or {}
        self.kwargs = kwargs

        # Enforce that we key on material_id
        self.materials.key = "material_id"
        self.thermo.key = "material_id"
        self.optimade.key = "material_id"

        super().__init__(sources=[materials, thermo], targets=optimade, **kwargs)

    def prechunk(self, number_splits: int) -> Iterator[Dict]:  # pragma: no cover
        """
        Prechunk method to perform chunking by the key field
        """
        q = dict(self.query)

        keys = self.optimade.newer_in(self.materials, criteria=q, exhaustive=True)

        N = ceil(len(keys) / number_splits)
        for split in grouper(keys, N):
            yield {"query": {self.materials.key: {"$in": list(split)}}}

    def get_items(self) -> Iterator:
        """
        Gets all items to process

        Returns:
            Generator or list of relevant materials
        """

        self.logger.info("Optimade Builder Started")

        q = dict(self.query)

        q.update({"deprecated": False})

        mat_ids = self.materials.distinct(self.materials.key, criteria=q)
        opti_ids = self.optimade.distinct(self.optimade.key)

        mats_set = set(
            self.optimade.newer_in(target=self.materials, criteria=q, exhaustive=True)
        ) | (set(mat_ids) - set(opti_ids))

        mats = [mat for mat in mats_set]

        self.total = len(mats)

        self.logger.info(f"Processing {self.total} items")

        for mat in mats:
            doc = self._get_processed_doc(mat)

            if doc is not None:
                yield doc
            else:
                pass

    def process_item(self, item):
        mpid = item["mat_doc"]["material_id"]
        structure = Structure.from_dict(item["mat_doc"]["structure"])
        last_updated_structure = item["mat_doc"]["last_updated"]

        # Functional names must be lowercase to adhere to optimade spec for querying attributes
        thermo_calcs = {}
        if item["thermo_doc"]:
            for doc in item["thermo_doc"]:
                thermo_calcs[doc["thermo_type"].lower()] = {
                    "thermo_id": doc["thermo_id"],
                    "energy_above_hull": doc["energy_above_hull"],
                    "formation_energy_per_atom": doc["formation_energy_per_atom"],
                    "last_updated_thermo": doc["last_updated"],
                }

        optimade_doc = OptimadeMaterialsDoc.from_structure(
            material_id=mpid,
            structure=structure,
            last_updated_structure=last_updated_structure,
            thermo_calcs=thermo_calcs,
        )

        doc = jsanitize(optimade_doc.model_dump(), allow_bson=True)

        return doc

    def update_targets(self, items):
        """
        Inserts the new optimade docs into the optimade collection
        """
        docs = list(filter(None, items))

        if len(docs) > 0:
            self.logger.info(f"Found {len(docs)} optimade docs to update")
            self.optimade.update(docs)
        else:
            self.logger.info("No items to update")

    def _get_processed_doc(self, mat):
        mat_doc = self.materials.query_one(
            {self.materials.key: mat}, [self.materials.key, "last_updated", "structure"]
        )

        mat_doc.update(
            {
                self.materials.key: mat_doc[self.materials.key],
            }
        )

        # Query thermo store for all docs matching material_id to catch
        # multiple stability calculations for the same material_id
        thermo_docs = self.thermo.query(
            {self.thermo.key: mat},
            [
                self.thermo.key,
                "thermo_type",
                "thermo_id",
                "energy_above_hull",
                "formation_energy_per_atom",
                "last_updated",
            ],
        )

        thermo_list = [doc for doc in thermo_docs]

        if thermo_list:
            for doc in thermo_list:
                doc.update({self.thermo.key: doc[self.thermo.key]})

        combined_doc = {
            "mat_doc": mat_doc,
            "thermo_doc": None if not thermo_list else thermo_list,
        }

        return combined_doc



from datetime import datetime
from itertools import chain
from math import ceil
from typing import Dict, Iterable, Iterator, List, Optional, Union

from maggma.builders import Builder
from maggma.stores import Store
from maggma.utils import grouper
from pymatgen.analysis.elasticity.strain import Deformation
from pymatgen.core.structure import Structure
from pymatgen.transformations.standard_transformations import (
    DeformStructureTransformation,
)

from emmet.builders.settings import EmmetBuildSettings
from emmet.core.utils import group_structures, jsanitize
from emmet.core.vasp.calc_types import TaskType
from emmet.core.vasp.material import MaterialsDoc
from emmet.core.vasp.task_valid import TaskDocument

__author__ = "Shyam Dwaraknath <shyamd@lbl.gov>"

SETTINGS = EmmetBuildSettings()


class MaterialsBuilder(Builder):
    """
    The Materials Builder matches VASP task documents by structure similarity into
    materials document. The purpose of this builder is group calculations and determine
    the best structure. All other properties are derived from other builders.

    The process is as follows:

        1.) Find all documents with the same formula
        2.) Select only task documents for the task_types we can select properties from
        3.) Aggregate task documents based on structure similarity
        4.) Create a MaterialDoc from the group of task documents
        5.) Validate material document

    """

    def __init__(
        self,
        tasks: Store,
        materials: Store,
        task_validation: Optional[Store] = None,
        query: Optional[Dict] = None,
        settings: Optional[EmmetBuildSettings] = None,
        **kwargs,
    ):
        """
        Args:
            tasks: Store of task documents
            materials: Store of materials documents to generate
            task_validation: Store for storing task validation results
            query: dictionary to limit tasks to be analyzed
            settings: EmmetSettings to use in the build process
        """

        self.tasks = tasks
        self.materials = materials
        self.task_validation = task_validation
        self.query = query if query else {}
        self.settings = EmmetBuildSettings.autoload(settings)
        self.kwargs = kwargs

        sources = [tasks]
        if self.task_validation:
            sources.append(self.task_validation)
        super().__init__(sources=sources, targets=[materials], **kwargs)

    def ensure_indexes(self):
        """
        Ensures indices on the tasks and materials collections
        """

        # Basic search index for tasks
        self.tasks.ensure_index("task_id")
        self.tasks.ensure_index("last_updated")
        self.tasks.ensure_index("state")
        self.tasks.ensure_index("formula_pretty")

        # Search index for materials
        self.materials.ensure_index("material_id")
        self.materials.ensure_index("last_updated")
        self.materials.ensure_index("task_ids")

        if self.task_validation:
            self.task_validation.ensure_index("task_id")
            self.task_validation.ensure_index("valid")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:  # pragma: no cover
        """Prechunk the materials builder for distributed computation"""
        temp_query = dict(self.query)
        temp_query["state"] = "successful"
        if len(self.settings.BUILD_TAGS) > 0 and len(self.settings.EXCLUDED_TAGS) > 0:
            temp_query["$and"] = [
                {"tags": {"$in": self.settings.BUILD_TAGS}},
                {"tags": {"$nin": self.settings.EXCLUDED_TAGS}},
            ]
        elif len(self.settings.BUILD_TAGS) > 0:
            temp_query["tags"] = {"$in": self.settings.BUILD_TAGS}

        self.logger.info("Finding tasks to process")
        all_tasks = list(
            self.tasks.query(temp_query, [self.tasks.key, "formula_pretty"])
        )

        processed_tasks = set(self.materials.distinct("task_ids"))
        to_process_tasks = {d[self.tasks.key] for d in all_tasks} - processed_tasks
        to_process_forms = {
            d["formula_pretty"]
            for d in all_tasks
            if d[self.tasks.key] in to_process_tasks
        }

        N = ceil(len(to_process_forms) / number_splits)

        for formula_chunk in grouper(to_process_forms, N):
            yield {"query": {"formula_pretty": {"$in": list(formula_chunk)}}}

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets all items to process into materials documents.
        This does no datetime checking; relying on whether
        task_ids are included in the Materials Collection

        Returns:
            generator or list relevant tasks and materials to process into materials
            documents
        """

        task_types = [t.value for t in self.settings.VASP_ALLOWED_VASP_TYPES]
        self.logger.info("Materials builder started")
        self.logger.info(f"Allowed task types: {task_types}")

        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Save timestamp to mark buildtime for material documents
        self.timestamp = datetime.utcnow()

        # Get all processed tasks:
        temp_query = dict(self.query)
        temp_query["state"] = "successful"
        if len(self.settings.BUILD_TAGS) > 0 and len(self.settings.EXCLUDED_TAGS) > 0:
            temp_query["$and"] = [
                {"tags": {"$in": self.settings.BUILD_TAGS}},
                {"tags": {"$nin": self.settings.EXCLUDED_TAGS}},
            ]
        elif len(self.settings.BUILD_TAGS) > 0:
            temp_query["tags"] = {"$in": self.settings.BUILD_TAGS}

        self.logger.info("Finding tasks to process")
        all_tasks = list(
            self.tasks.query(temp_query, [self.tasks.key, "formula_pretty"])
        )

        processed_tasks = set(self.materials.distinct("task_ids"))
        to_process_tasks = {d[self.tasks.key] for d in all_tasks} - processed_tasks
        to_process_forms = {
            d["formula_pretty"]
            for d in all_tasks
            if d[self.tasks.key] in to_process_tasks
        }

        self.logger.info(f"Found {len(to_process_tasks)} unprocessed tasks")
        self.logger.info(f"Found {len(to_process_forms)} unprocessed formulas")

        # Set total for builder bars to have a total
        self.total = len(to_process_forms)

        if self.task_validation:
            invalid_ids = {
                doc[self.tasks.key]
                for doc in self.task_validation.query(
                    {"valid": False}, [self.task_validation.key]
                )
            }
        else:
            invalid_ids = set()

        projected_fields = [
            "last_updated",
            "completed_at",
            "task_id",
            "formula_pretty",
            "output.energy_per_atom",
            "output.structure",
            "input.parameters",
            # needed for run_type and task_type
            "calcs_reversed.input.parameters",
            "calcs_reversed.input.incar",
            "orig_inputs",
            "input.structure",
            # needed for entry from task_doc
            "output.energy",
            "input.is_hubbard",
            "input.hubbards",
            "input.potcar_spec",
            # needed for transform deformation structure back for grouping
            "transformations",
            # misc info for materials doc
            "tags",
        ]

        for formula in to_process_forms:
            tasks_query = dict(temp_query)
            tasks_query["formula_pretty"] = formula
            tasks = list(
                self.tasks.query(criteria=tasks_query, properties=projected_fields)
            )
            for t in tasks:
                t["is_valid"] = t[self.tasks.key] not in invalid_ids

            yield tasks

    def process_item(self, items: List[Dict]) -> List[Dict]:
        """
        Process the tasks into a list of materials

        Args:
            tasks [dict]: a list of task docs

        Returns:
            ([dict],list): a list of new materials docs and a list of task_ids that
                were processed
        """

        tasks = [TaskDocument(**task) for task in items]
        formula = tasks[0].formula_pretty
        task_ids = [task.task_id for task in tasks]

        # not all tasks contains transformation information
        task_transformations = [task.get("transformations", None) for task in items]

        self.logger.debug(f"Processing {formula}: {task_ids}")

        grouped_tasks = self.filter_and_group_tasks(tasks, task_transformations)
        materials = []
        for group in grouped_tasks:
            try:
                materials.append(
                    MaterialsDoc.from_tasks(
                        group,
                        structure_quality_scores=self.settings.VASP_STRUCTURE_QUALITY_SCORES,
                        use_statics=self.settings.VASP_USE_STATICS,
                    )
                )
            except Exception as e:
                failed_ids = list({t_.task_id for t_ in group})
                doc = MaterialsDoc.construct_deprecated_material(group)
                doc.warnings.append(str(e))
                materials.append(doc)
                self.logger.warn(
                    f"Failed making material for {failed_ids}."
                    f" Inserted as deprecated Material: {doc.material_id}"
                )

        self.logger.debug(f"Produced {len(materials)} materials for {formula}")

        return jsanitize([mat.model_dump() for mat in materials], allow_bson=True)

    def update_targets(self, items: List[List[Dict]]):
        """
        Inserts the new task_types into the task_types collection

        Args:
            items ([([dict],[int])]): A list of tuples of materials to update and the
                corresponding processed task_ids
        """

        docs = list(chain.from_iterable(items))  # type: ignore

        for item in docs:
            item.update({"_bt": self.timestamp})

        material_ids = list({item["material_id"] for item in docs})

        if len(items) > 0:
            self.logger.info(f"Updating {len(docs)} materials")
            self.materials.remove_docs({self.materials.key: {"$in": material_ids}})
            self.materials.update(docs=docs, key=["material_id"])
        else:
            self.logger.info("No items to update")

    def filter_and_group_tasks(
        self, tasks: List[TaskDocument], task_transformations: List[Union[Dict, None]]
    ) -> Iterator[List[TaskDocument]]:
        """
        Groups tasks by structure matching
        """

        filtered_tasks = []
        filtered_transformations = []
        for task, transformations in zip(tasks, task_transformations):
            if any(
                allowed_type == task.task_type
                for allowed_type in self.settings.VASP_ALLOWED_VASP_TYPES
            ):
                filtered_tasks.append(task)
                filtered_transformations.append(transformations)

        structures = []
        for idx, (task, transformations) in enumerate(
            zip(filtered_tasks, filtered_transformations)
        ):
            if task.task_type == TaskType.Deformation:
                if (
                    transformations is None
                ):  # Do not include deformed tasks without transformation information
                    self.logger.debug(
                        "Cannot find transformation for deformation task {}. Excluding task.".format(
                            task.task_id
                        )
                    )
                    continue
                else:
                    s = undeform_structure(task.input.structure, transformations)

            else:
                s = task.output.structure
            s.index: int = idx  # type: ignore
            structures.append(s)

        grouped_structures = group_structures(
            structures,
            ltol=self.settings.LTOL,
            stol=self.settings.STOL,
            angle_tol=self.settings.ANGLE_TOL,
            symprec=self.settings.SYMPREC,
        )
        for group in grouped_structures:
            grouped_tasks = [filtered_tasks[struct.index] for struct in group]  # type: ignore
            yield grouped_tasks


def undeform_structure(structure: Structure, transformations: Dict) -> Structure:
    """
    Get the undeformed structure by applying the transformations in a reverse order.

    Args:
        structure: deformed structure
        transformation: transformation that deforms the structure

    Returns:
        undeformed structure
    """

    for transformation in reversed(transformations.get("history", [])):
        if transformation["@class"] == "DeformStructureTransformation":
            deform = Deformation(transformation["deformation"])
            dst = DeformStructureTransformation(deform.inv)
            structure = dst.apply_transformation(structure)
        else:
            raise RuntimeError(
                "Expect transformation to be `DeformStructureTransformation`; "
                f"got {transformation['@class']}"
            )

    return structure



from importlib.metadata import version
from typing import TYPE_CHECKING, Optional, Union

from maggma.builders.map_builder import MapBuilder
from maggma.core import Store
from matcalc.util import get_universal_calculator
from pymatgen.core import Structure

from emmet.core.ml import MLDoc
from emmet.core.utils import jsanitize

if TYPE_CHECKING:
    from ase.calculators.calculator import Calculator


class MLBuilder(MapBuilder):
    def __init__(
        self,
        materials: Store,
        ml_potential: Store,
        model: Union[str, "Calculator"],
        model_kwargs: Optional[dict] = None,
        prop_kwargs: Optional[dict] = None,
        provenance: Optional[dict] = None,
        **kwargs
    ):
        """Machine learning interatomic potential builder.

        Args:
            materials (Store): Materials to use as input structures.
            ml_potential (Store): Where to save MLDoc documents to.
            model (str | Calculator): ASE calculator or name of model to use as ML
                potential. See matcalc.util.UNIVERSAL_CALCULATORS for recognized names.
            model_kwargs (dict, optional): Additional kwargs to pass to the calculator.
                Defaults to None.
            prop_kwargs (dict[str, dict], optional): Separate kwargs passed to each matcalc
                PropCalc class. Recognized keys are RelaxCalc, ElasticityCalc, PhononCalc, EOSCalc.
                Defaults to None.
            provenance (dict, optional): Additional provenance information to include in
                MLDocs. Will be saved in each document so use sparingly. Defaults to None.
                Set to {} to disable default provenance model_name, model_version, matcalc_version.
        """
        self.materials = materials
        self.ml_potential = ml_potential
        self.kwargs = kwargs
        self.model = get_universal_calculator(model, **(model_kwargs or {}))
        self.prop_kwargs = prop_kwargs or {}

        if provenance == {}:
            self.provenance = {}
        else:
            model_name = (
                model if isinstance(model, str) else type(model).__name__
            ).lower()
            model_name = {"chgnetcalculator": "chgnet"}.get(model_name, model_name)
            pkg_name = {"m3gnet": "matgl"}.get(model_name, model_name)
            self.provenance = dict(
                model_name=model_name,
                model_version=version(pkg_name),
                matcalc_version=version("matcalc"),
                **(provenance or {}),
            )

        # Enforce that we key on material_id
        self.materials.key = "material_id"
        self.ml_potential.key = "material_id"
        super().__init__(
            source=materials,
            target=ml_potential,
            projection=["structure", "deprecated"],
            **kwargs,
        )

    def unary_function(self, item):
        struct = Structure.from_dict(item["structure"])
        mp_id, deprecated = item["material_id"], item["deprecated"]

        doc = MLDoc(
            structure=struct,
            material_id=mp_id,
            calculator=self.model,
            prop_kwargs=self.prop_kwargs,
            deprecated=deprecated,
            **self.provenance,
        )

        return jsanitize(doc, allow_bson=True)
